import { mapSerializer, base58, SdkError } from '@metaplex-foundation/umi';
import { shortU16 } from '@metaplex-foundation/umi-serializer-data-view';
import { toWeb3JsMessageFromInput, fromWeb3JsMessage } from '@metaplex-foundation/umi-web3js-adapters';
import { VersionedTransaction } from '@solana/web3.js';

/* eslint-disable no-bitwise */
const TRANSACTION_VERSION_FLAG = 0x80;
const TRANSACTION_VERSION_MASK = 0x7f;
function createWeb3JsTransactionFactory(context) {
  const create = input => {
    const web3JsMessage = toWeb3JsMessageFromInput(input);
    const message = fromWeb3JsMessage(web3JsMessage);
    const web3JsTransaction = new VersionedTransaction(web3JsMessage, input.signatures);
    return {
      message,
      serializedMessage: serializeMessage(message),
      signatures: web3JsTransaction.signatures
    };
  };
  const serialize = transaction => getTransactionSerializer().serialize(transaction);
  const deserialize = serializedTransaction => getTransactionSerializer().deserialize(serializedTransaction)[0];
  const serializeMessage = message => getTransactionMessageSerializer().serialize(message);
  const deserializeMessage = serializedMessage => getTransactionMessageSerializer().deserialize(serializedMessage)[0];
  const getTransactionSerializer = () => {
    const s = context.serializer;
    return {
      ...mapSerializer(s.struct([['signatures', s.array(s.bytes({
        size: 64
      }), {
        size: shortU16()
      })], ['serializedMessage', s.bytes()]]), value => value, value => ({
        ...value,
        message: deserializeMessage(value.serializedMessage)
      })),
      description: 'Transaction'
    };
  };
  const getTransactionMessageSerializer = () => ({
    description: 'TransactionMessage',
    fixedSize: null,
    maxSize: null,
    serialize: value => {
      const serializer = getTransactionMessageSerializerForVersion(value.version);
      return serializer.serialize(value);
    },
    deserialize: (bytes, offset = 0) => {
      const [version] = getTransactionVersionSerializer().deserialize(bytes, offset);
      const serializer = getTransactionMessageSerializerForVersion(version);
      return serializer.deserialize(bytes, offset);
    }
  });
  const getTransactionMessageSerializerForVersion = version => {
    const s = context.serializer;
    return s.struct([['version', getTransactionVersionSerializer()], ['header', getTransactionMessageHeaderSerializer()], ['accounts', s.array(s.publicKey(), {
      size: shortU16()
    })], ['blockhash', s.string({
      encoding: base58,
      size: 32
    })], ['instructions', s.array(getCompiledInstructionSerializer(), {
      size: shortU16()
    })], ['addressLookupTables', s.array(getCompiledAddressLookupTableSerializer(), {
      size: version === 'legacy' ? 0 : shortU16()
    })]]);
  };
  const getTransactionVersionSerializer = () => ({
    description: 'TransactionVersion',
    fixedSize: null,
    maxSize: 1,
    serialize: value => {
      if (value === 'legacy') return new Uint8Array([]);
      return new Uint8Array([TRANSACTION_VERSION_FLAG | value]);
    },
    deserialize: (bytes, offset = 0) => {
      const slice = bytes.slice(offset);
      if (slice.length === 0 || (slice[0] & TRANSACTION_VERSION_FLAG) === 0) {
        return ['legacy', offset];
      }
      const version = slice[0] & TRANSACTION_VERSION_MASK;
      if (version > 0) {
        throw new SdkError(`Unsupported transaction version: ${version}.`);
      }
      return [version, offset + 1];
    }
  });
  const getTransactionMessageHeaderSerializer = () => {
    const s = context.serializer;
    return s.struct([['numRequiredSignatures', s.u8()], ['numReadonlySignedAccounts', s.u8()], ['numReadonlyUnsignedAccounts', s.u8()]]);
  };
  const getCompiledInstructionSerializer = () => {
    const s = context.serializer;
    return s.struct([['programIndex', s.u8()], ['accountIndexes', s.array(s.u8(), {
      size: shortU16()
    })], ['data', s.bytes({
      size: shortU16()
    })]]);
  };
  const getCompiledAddressLookupTableSerializer = () => {
    const s = context.serializer;
    return s.struct([['publicKey', s.publicKey()], ['writableIndexes', s.array(s.u8(), {
      size: shortU16()
    })], ['readonlyIndexes', s.array(s.u8(), {
      size: shortU16()
    })]]);
  };
  return {
    create,
    serialize,
    deserialize,
    serializeMessage,
    deserializeMessage
  };
}

export { createWeb3JsTransactionFactory };
//# sourceMappingURL=createWeb3JsTransactionFactory.mjs.map
